/*
    Interval: Interval.m
    Copyright (C) 1999-2003  Macaulay Institute

    This file is part of Interval, a class to implement machine interval
    arithmetic.

    Interval is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Interval is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/

#import "Interval.h"
#import <ieeefp.h>
#import <stdlib.h>
#import <errno.h>
#import <float.h>
#import <unistd.h>
#import <math.h>
//#import <values.h>
#import <defobj.h>

/*****************************************************************************/

/* Typedefs */

/* STACK
 *
 * Typedef for stack to contain intermediate variables in
 * expressions. These are created by methods like sum: and diff:, and
 * then popped off by methods like print and set.
 */

typedef struct stack {
  Interval *top;
  struct stack *next;
} STACK;

/* PSTACK
 *
 * Typedef for stack to contain strings created by the printf:
 * method. These are freed when the method printfclean is called.
 */

typedef struct print_stack {
  char *str;
  struct print_stack *next;
} PSTACK;

/*****************************************************************************/

/* Macros */

#define TRACE
#undef TRACE			// Comment out to do a trace

#define RZ save_rnd = fpsetround(FP_RZ)
#define RN save_rnd = fpsetround(FP_RN)
#define RP save_rnd = fpsetround(FP_RP)
#define RM save_rnd = fpsetround(FP_RM)
				// Shorthand for all the various rounding modes
				// that save the original mode
#define RS fpsetround(save_rnd)	// Restore the saved rounding mode

#define IF_STACK_TOP(arg) if(stack != NULL && stack->top == (arg))
#define STACK_TOP(arg) (stack != NULL && stack->top == (arg))
#define IF_NOT_STACK_TOP(arg) if(stack == NULL || stack->top != (arg))

#define CHECK_DEF warn_defined(self, _cmd)
#define CHECK_DEF_ARG(x) warn_defined(self, _cmd); warn_defined((x), _cmd)

#define TRACE_SELF if(trace) { printf("TRACE: %s (Method %s): ", \
                                      trace, sel_get_name(_cmd)); \
                               [self bufwrite]; \
                               printf("\n"); }
#define TRACE_SELF_ARG(x) if(trace) { \
                            printf("TRACE: %s (Method %s ", \
                                   trace, sel_get_name(_cmd)); \
                            if(x) [(x) bufwrite]; else printf("nil"); \
                            printf("): "); \
                            [self bufwrite]; \
                            printf("\n"); }

#define CHECK_CONST if(defined && constant) { \
                       fprintf(stderr, "FATAL: Changing constant interval\n");\
                       abort(); }

#define CHECK_NEG if(negative) { \
                    if(max > 0.0) max = nextafter(0.0, -1.0); \
                    force_interval(&var, &min, &max, posneg_crash, max); }

#define CHECK_POS if(positive) { \
                    if(min < 0.0) min = nextafter(0.0, 1.0); \
                    force_interval(&var, &min, &max, posneg_crash, min); }

/* BINARY_OP
 *
 * Macro to do a binary infix operator in interval arithmetic. This
 * involves finding the minimum and maximum values of each combination
 * of the minimum and maximum of the input intervals [min1, max1] and
 * [min2, max2], rounded towards positive and negative infinity. The
 * whole exercise has to be conducted with caution because amin could
 * be the same variable as min1, as could amax and max1. The first
 * thing to do, therefore, is to start a new statement block with its
 * own local variables (so we're not messing with any variables in the
 * enclosing block) to contain the eight possible results of operating
 * on the two ranges. A series of comparisons is then used to
 * determine which of these results is the minimum and maximum to use
 * for the answer. After that, aval is computed as val1 op val2 in
 * round-to-nearest mode, before restoring the rounding mode to its
 * original value. This macro is used by the mul: and div: operators.
 */

#define BINARY_OP(op, amin, amax, aval, min1, max1, val1, min2, max2, val2) \
  { \
    double amin1, amin2, amin3, amin4; \
    double amax1, amax2, amax3, amax4; \
    save_rnd = fpsetround(FP_RM); \
    amin1 = (min1) op (min2); \
    amin2 = (min1) op (max2); \
    amin3 = (max1) op (min2); \
    amin4 = (max1) op (max2); \
    fpsetround(FP_RP); \
    amax1 = (max1) op (max2); \
    amax2 = (max1) op (min2); \
    amax3 = (min1) op (max2); \
    amax4 = (min1) op (min2); \
    if(amin4 > amin3) { \
      if(amin2 > amin1) { \
        if(amin3 > amin1) (amin) = amin1; \
        else (amin) = amin3; \
      } \
      else { \
        if(amin3 > amin2) (amin) = amin2; \
        else (amin) = amin3; \
      } \
    } \
    else { \
      if(amin2 > amin1) { \
        if(amin4 > amin1) (amin) = amin1; \
        else (amin) = amin4; \
      } \
      else { \
        if(amin4 > amin2) (amin) = amin2; \
        else (amin) = amin4; \
      } \
    } \
    if(amax4 < amax3) { \
      if(amax2 < amax1) { \
        if(amax3 < amax1) (amax) = amax1; \
        else (amax) = amax3; \
      } \
      else { \
        if(amax3 < amax2) (amax) = amax2; \
        else (amax) = amax3; \
      } \
    } \
    else { \
      if(amax2 < amax1) { \
        if(amax4 < amax1) (amax) = amax1; \
        else (amax) = amax4; \
      } \
      else { \
        if(amax4 < amax2) (amax) = amax2; \
        else (amax) = amax4; \
      } \
    } \
  } \
  fpsetround(FP_RN); \
  (aval) = (val1) op (val2); \
  fpsetround(save_rnd)

/* UNARY_PREFIX_OP
 *
 * Just as for BINARY_OP, try all the ways of applying op to the input
 * range [min1, max1] rounding towards positive and negative infinity,
 * and taking the minimum and maximum result.
 */
  
#define UNARY_PREFIX_OP(op, amin, amax, aval, min1, max1, val1) \
  { \
    double amin1, amin2, amax1, amax2; \
    save_rnd = fpsetround(FP_RM); \
    amin1 = op (min1); \
    amin2 = op (max1); \
    fpsetround(FP_RP); \
    amax1 = op (max1); \
    amax2 = op (min1); \
    if(amin1 > amin2) amin = amin2; \
    else amin = amin1; \
    if(amax1 < amax2) amax = amax2; \
    else amax = amax1; \
  } \
  fpsetround(FP_RN); \
  (aval) = op (val1); \
  fpsetround(save_rnd)

/* TWO_ARG_FUNC
 *
 * The same as BINARY_OP, but using a function call instead of an
 * operator.
 */

#define TWO_ARG_FUNC(f, amin, amax, aval, min1, max1, val1, min2, max2, val2) \
  { \
    double amin1, amin2, amin3, amin4; \
    double amax1, amax2, amax3, amax4; \
    save_rnd = fpsetround(FP_RM); \
    amin1 = f((min1), (min2)); \
    amin2 = f((min1), (max2)); \
    amin3 = f((max1), (min2)); \
    amin4 = f((max1), (max2)); \
    fpsetround(FP_RP); \
    amax1 = f((max1), (max2)); \
    amax2 = f((max1), (min2)); \
    amax3 = f((min1), (max2)); \
    amax4 = f((min1), (min2)); \
    if(amin4 > amin3) { \
      if(amin2 > amin1) { \
        if(amin3 > amin1) (amin) = amin1; \
        else (amin) = amin3; \
      } \
      else { \
        if(amin3 > amin2) (amin) = amin2; \
        else (amin) = amin3; \
      } \
    } \
    else { \
      if(amin2 > amin1) { \
        if(amin4 > amin1) (amin) = amin1; \
        else (amin) = amin4; \
      } \
      else { \
        if(amin4 > amin2) (amin) = amin2; \
        else (amin) = amin4; \
      } \
    } \
    if(amax4 < amax3) { \
      if(amax2 < amax1) { \
        if(amax3 < amax1) (amax) = amax1; \
        else (amax) = amax3; \
      } \
      else { \
        if(amax3 < amax2) (amax) = amax2; \
        else (amax) = amax3; \
      } \
    } \
    else { \
      if(amax2 < amax1) { \
        if(amax4 < amax1) (amax) = amax1; \
        else (amax) = amax4; \
      } \
      else { \
        if(amax4 < amax2) (amax) = amax2; \
        else (amax) = amax4; \
      } \
    } \
  } \
  fpsetround(FP_RN); \
  (aval) = (val1) op (val2); \
  fpsetround(save_rnd)

/* ONE_ARG_FUNC
 *
 * The same as UNARY_PREFIX_OP, but with a function call instead of
 * the operator.
 */
  
#define ONE_ARG_FUNC(f, amin, amax, aval, min1, max1, val1) \
  { \
    double amin1, amax1, amin2, amax2; \
    save_rnd = fpsetround(FP_RM); \
    amin1 = f((min1)); \
    amin2 = f((max1)); \
    fpsetround(FP_RP); \
    amax1 = f((max1)); \
    amax2 = f((min1)); \
    if(amin1 > amin2) amin = amin2; \
    else amin = amin1; \
    if(amax1 < amax2) amax = amax2; \
    else amax = amax1; \
  } \
  fpsetround(FP_RN); \
  (aval) = op (val1); \
  fpsetround(save_rnd)

/*****************************************************************************/

/* Global variables */

static STACK *stack = NULL;	// The intermediate Intervals stack
static unsigned long stack_size = 0LU;
				// The number of items on this stack
static unsigned long nintervals = 0LU;
				// The number of Intervals that have been
				// created but not dropped

static PSTACK *pstack = NULL;	// The printf strings stack

static fp_rnd save_rnd = FP_RN; // The rounding mode used externally
				// to the class

/* Class variables used to determine options */

static BOOL nan_crash = NO;	// Whether to crash if NaN or Inf is known to
				// be a possible result of the calculation
static BOOL strict_div = YES;	// Whether to be strict about divisors
				// containing zero
static BOOL hide_warn = NO;	// Whether to hide warning messages

static BOOL undef_crash = NO;	// Whether to crash if an operation is done
				// on an undefined interval
static BOOL posneg_crash = NO;	// Whether to crash if the original variable
				// (var) has the wrong sign in a mathematically
				// proven signed interval

/*****************************************************************************/

/* Static functions */

/* pop() -> Interval
 *
 * Pop an Interval off the intermediate Intervals stack and return it
 */

static Interval *pop(void) {
  if(stack == NULL) {		// Attempt to pop off an empty stack!
    fprintf(stderr, "Interval stack underflow\n");
    abort();
  }
  else {
    STACK *tmp = stack;
    Interval *item = stack->top;
#   ifdef TRACE
    double min, var, max;

    [item get: &var min: &min max: &max];
    printf("Popped %p [%g, %g, %g]\n", item, min, var, max);
#   endif

    stack = stack->next;
    free(tmp);
    stack_size--;
    return item;
  }
}

/* push() -> Interval
 *
 * Push an Interval on to the intermediate Intervals stack and return it.
 */

static Interval *push(Interval *item) {
  STACK *tmp = stack;
# ifdef TRACE
  double min, var, max;

  [item get: &var min: &min max: &max];
  printf("Pushed %p [%g, %g, %g]\n", item, min, var, max);
# endif

  stack = (STACK *)malloc(sizeof(STACK));
  if(stack == NULL) {
    perror("Interval stack overflow");
    abort();
  }
  stack->top = item;
  stack->next = tmp;

  stack_size++;

  return item;
}

/* check_stack(lhs, rhs)
 *
 * This function is called by a number of methods that need to check
 * whether self or the argument to the method is on the stack. If the
 * argument is on the stack, it needs to be dropped as it will not be
 * used any further. If self is on the stack, then it needs to stay on
 * the stack if the method returns an Interval containing the
 * answer. The order in which items appear on the stack is
 * assumed. e.g. in [[a] op: [b]], the result of [a] will be on the
 * stack first, followed by the result of [b]. (This has been checked
 * in Cygwin and Solaris, but may not necessarily apply elsewhere.)
 * Thus [b] is on the top of the stack, followed by [a]. (Doh!)
 */

// static void check_stack(Interval *lhs, Interval *rhs) {
//   IF_STACK_TOP(rhs) [pop() drop];

  /* Code if [a] is on top of the stack, followed by [b]: */
//   IF_STACK_TOP(lhs) {
//     pop();
//     IF_STACK_TOP(rhs) [pop() drop];
//     push(lhs);
//   }
//   else IF_STACK_TOP(rhs) [pop() drop];
// }


/* clear_pstack()
 *
 * Called from the printfclean method to clean up strings created by a
 * printf: method call.
 */

static void clear_pstack(void) {
  while(pstack != NULL) {
    PSTACK *tmp = pstack;

    pstack = pstack->next;
    free(tmp->str);
    free(tmp);
  }
}

/* push_pstack(s)
 *
 * Called from the printf: method to store the string created so it
 * can be freed later.
 */

static void push_pstack(char *s) {
  PSTACK *tmp = pstack;

  pstack = (PSTACK *)malloc(sizeof(PSTACK));
  if(pstack == NULL) {
    perror("Memory allocation");
    abort();
  }
  pstack->str = s;
  pstack->next = tmp;
}


/* warn_uncertainty(lhs, rhs, str)
 *
 * Called from the comparison operator methods to provide a
 * standardised warning message format when there is uncertainty in
 * the truth value of a comparison due to an overlap between the range
 * defined by lhs and that of rhs.
 */

static void warn_uncertainty(Interval *lhs, Interval *rhs, const char *str) {
  if(!hide_warn) {
    fprintf(stderr, "WARNING: Uncertainty comparing ");
    [lhs bufwrite: stderr];
    fprintf(stderr, " and ");
    [rhs bufwrite: stderr];
    fprintf(stderr, ": %s\n", str);
  }
}

/* warn_defined(obj, caller)
 *
 * Check whether an interval is defined and issue an appropriate warning.
 */

static void warn_defined(Interval *obj, SEL caller) {
  if(!hide_warn) {
    if(!obj->defined) {
      if(undef_crash) fprintf(stderr, "FATAL: ");
      else fprintf(stderr, "WARNING: ");
      fprintf(stderr, "Attempt to operate on undefined Interval %p "
	      "(method %s)\n", obj, sel_get_name(caller));
      if(undef_crash) abort();
    }
  }
}

/* force_interval(var, min, max, crash)
 *
 * Force the interval to a legal state. If crash is YES, then die if var
 * has to be changed. It is assumed that the only reason the interval is
 * illegal is because of a change to either min or max.
 */

void force_interval(double *var, double *min, double *max, BOOL crash,
		    double limit) {
  if((*max) < (*var)) {
    if(crash) {
      fprintf(stderr, "FATAL: Forced change to original variable\n");
      abort();
    }
    (*var) = (*max);
  }
  if((*min) > (*var)) {
    if(crash) {
      fprintf(stderr, "FATAL: Forced change to original variable\n");
      abort();
    }
    (*var) = (*min);
  }
  if((*min) > (*max)) (*min) = (*max) = (*var) = limit;
}

/* necessarily_OP(lmin, lmax, rmin, rmax)
 *
 * For each comparison operator OP, a function is defined below for
 * the sake of standardisation throughout the class. This defines the
 * modal comparison operators in "necessarily" mode for the ranges L =
 * [lmin, lmax] and R = [rmin, rmax], in which the result is true iff
 * OP holds for all members of L and R.
 */

static BOOL necessarily_eq(double lmin, double lmax,
			   double rmin, double rmax) {
  return (lmin == rmax && rmin == lmax && lmin == lmax) ? YES : NO;
}

static BOOL necessarily_ne(double lmin, double lmax,
			   double rmin, double rmax) {
  return (lmin > rmax || lmax < rmin) ? YES : NO;
}

static BOOL necessarily_gt(double lmin, double lmax,
			   double rmin, double rmax) {
  return (lmin > rmax) ? YES : NO;
}

static BOOL necessarily_ge(double lmin, double lmax,
			   double rmin, double rmax) {
  return (lmin >= rmax) ? YES : NO;
}

static BOOL necessarily_lt(double lmin, double lmax,
			   double rmin, double rmax) {
  return (lmax < rmin) ? YES : NO;
}

static BOOL necessarily_le(double lmin, double lmax,
			   double rmin, double rmax) {
  return (lmax <= rmin) ? YES : NO;
}

/* possibly_OP(lmin, lmax, rmin, rmax)
 *
 * For each comparison operator OP, a function is defined below for
 * the sake of standardisation throughout the class. This defines the
 * modal comparison operators in "possibly" mode for the ranges L =
 * [lmin, lmax] and R = [rmin, rmax], in which the result is true iff
 * there exists a member of L, l, and a member of R, r, for which l OP
 * r is true. Modal logic is defined such that possibly(not(X)) <=>
 * not(necessarily(X)), meaning that the "possibly" mode operators can
 * all be defined in terms of the "necesssarily" mode ones. It is
 * assumed that eq <=> not(ne), gt <=> not(le), and lt <=> not(ge).
 */

static BOOL possibly_eq(double lmin, double lmax, double rmin, double rmax) {
  return !necessarily_ne(lmin, lmax, rmin, rmax);
}

static BOOL possibly_ne(double lmin, double lmax, double rmin, double rmax) {
  return !necessarily_eq(lmin, lmax, rmin, rmax);
}

static BOOL possibly_gt(double lmin, double lmax, double rmin, double rmax) {
  return !necessarily_le(lmin, lmax, rmin, rmax);
}

static BOOL possibly_ge(double lmin, double lmax, double rmin, double rmax) {
  return !necessarily_lt(lmin, lmax, rmin, rmax);
}

static BOOL possibly_lt(double lmin, double lmax, double rmin, double rmax) {
  return !necessarily_ge(lmin, lmax, rmin, rmax);
}

static BOOL possibly_le(double lmin, double lmax, double rmin, double rmax) {
  return !necessarily_gt(lmin, lmax, rmin, rmax);
}

/* minsigdig(number) -> number of digits
 *
 * Return the minimum number of digits to print after the decimal
 * point in a %.*e format so that the argument is restored in round to
 * nearest mode.
 */

int minsigdig(double number) {
  int nsigdig = 0;
  double recovernum;
  fp_rnd cur_dir;

  if(isnan(number)) return 1;	// Deal with NaNs first.
  cur_dir = fpsetround(FP_RN);
  do {
    char *buf;

    nsigdig++;
    buf = (char *)malloc((nsigdig + 11) * sizeof(char));
				// Allow for -N.___E-dddd\0, where ___ is
				// replaced with nsigdig digits, plus one
				// extra for luck!
    if(buf == NULL) {
      perror("Memory allocation");
      abort();
    }

    sprintf(buf, "%.*e", nsigdig, number);
    sscanf(buf, "%lf", &recovernum);
    free(buf);
  } while(recovernum != number);

  fpsetround(cur_dir);
  return nsigdig;
}


/* interval_zone() -> Zone
 *
 * Intervals have their own zone, so users don't have to worry about
 * them. This is why the create method does not take any
 * arguments. However, the create method needs a zone to give to the
 * superclass when allocating itself. This function's job is to
 * provide that zone. The first time it is called, it creates a zone
 * before returning it, thereafter, it returns the zone. Thus wherever
 * one would normally write a zone in the code, it should be replaced
 * with a call to this function.  */

static id <Zone> interval_zone(void) {
  static id <Zone> zone = nil;
  static BOOL zone_defined = NO;

  if(!zone_defined) {
    zone = [Zone create: globalZone];
    zone_defined = YES;
  }
  return zone;
}

/*****************************************************************************/

/* Private method declarations */

@interface Interval (PrivateInterval)

+createWith: (double)value defined: (BOOL)def
   positive: (BOOL)pos negative: (BOOL)neg constant: (BOOL)con;

+createWith: (double)value min: (double)lo max: (double)hi defined: (BOOL)def
   positive: (BOOL)pos negative: (BOOL)neg constant: (BOOL)con;

@end

/* Implementation of the Interval class */

@implementation Interval

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Class methods for accessing Interval-related objects and information */

/* stackSize -> size
 *
 * Return the current size of the intermediate result stack. This is
 * useful for debugging purposes, as, outside an expression, it should
 * always be zero.
 */

+(unsigned long)stackSize {
  return stack_size;
}

/* intervalCount -> count
 *
 * Return the current number of active Intervals -- those created but
 * not dropped. Another method useful for debugging purposes, and a
 * potentially useful way of checking for memory leaks.
 */

+(unsigned long)intervalCount {
  return nintervals;
}

/* getZone -> interval_zone
 *
 * Return the interval zone.
 */

+getZone {
  return interval_zone();
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Methods setting options for the behaviour and strictness of Intervals */

/* showWarnings
 *
 * Set the flag that causes warning messages to be sent to standard error.
 */

+showWarnings {
  hide_warn = NO;
  return self;
}

/* hideWarnings
 *
 * Set the flag that causes warning messages to be ignored.
 */

+hideWarnings {
  hide_warn = YES;
  return self;
}

/* fatalNaN
 *
 * If NaN or Inf is known to be the result of a calculation, then the
 * program will abort.
 */

+fatalNaN {
  nan_crash = YES;
  return self;
}

/* ignoreNaN
 *
 * Ignore cases when NaN or Inf could be the result of a calculation
 */

+ignoreNaN {
  nan_crash = NO;
  return self;
}

/* strictDivZero
 *
 * If the divisor range contains zero, then cause the result to contain
 * infinity or NaN if the dividend also contains zero.
 */

+strictDivZero {
  strict_div = YES;
  return self;
}

/* ignoreDivZero
 *
 * If the divisor range contains zero, then just keep going using the ranges
 * as specified. Results are potentially rubbish.
 */

+ignoreDivZero {
  strict_div = NO;
  fprintf(stderr, "WARNING: Divisions by ranges containing zero will not "
	  "be treated properly. Results are potentially nonsense!\n");
  return self;
}

/* fatalUndef
 *
 * If an operation is performed using an undefined variable, then die.
 */

+fatalUndef {
  undef_crash = YES;
  return self;
}

/* ignoreUndef
 *
 * If an operation is performed using an undefined variable, then care not.
 */

+ignoreUndef {
  undef_crash = NO;
  return self;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Various methods returning a newly created Interval object */

/* constant: value -> Interval [PUSH]
 *
 * This method should be used to replace any double constants
 * appearing in an expression. It pushes the newly created constant
 * onto the stack.
 */

+constant: (double)value {
  return push([self createWith: value]);
}

/* castInt: value -> Interval [PUSH]
 *
 * If an integer appears as part of a double expression, it needs to
 * be cast into a constant and pushed onto the stack much as per the
 * constant: method.
 */

+castInt: (int)value {
  double cmin, cmax, cval;

  UNARY_PREFIX_OP((double), cmin, cmax, cval, value, value, value);
  return push([self createWith: cval min: cmin max: cmax]);
}

/* castULL: value -> Interval [PUSH]
 *
 * If an unsigned long long appears as part of a double expression, it
 * needs to be cast into a constant and pushed onto the stack much as
 * per the constant: method.
 */

+castULL: (unsigned long long)value {
  double cmin, cmax, cval;

  UNARY_PREFIX_OP((double), cmin, cmax, cval, value, value, value);
  return push([self createWith: cval min: cmin max: cmax]);
}

/* create -> Interval
 *
 * Allocate memory for a new Interval, and assign it a default value.
 *
 * <private>Should this method assign a default value of "unassigned"
 * and check this when used later in the code? It would involve a
 * rather inconvenient overhead, but would also be more
 * rigorous.</private>
 */

+create {
  return [self createWith: 0.0 defined: NO
	       positive: NO negative: NO constant: NO];
}

/* createWith: value -> Interval
 *
 * Allocate memory for a new Interval and assign it the given value.
 */

+createWith: (double)value {
  return [self createWith: value defined: YES
	       positive: NO negative: NO constant: NO];
}

/* createWith: value min: lo max: hi -> Interval
 *
 * Creation method with a specified range.
 */

+createWith: (double)value min: (double)lo max: (double)hi {
  return [self createWith: value min: lo max: hi defined: YES
	       positive: NO negative: NO constant: NO];
}

/* createPositive -> Interval
 *
 * Create an Interval that has been (independently) mathematically proven to
 * only ever take positive values.
 */

+createPositive {
  return [self createWith: nextafter(0.0, 1.0) defined: NO
	       positive: YES negative: NO constant: NO];
}

/* createPositiveWith: value -> Interval
 *
 * Create an Interval that has been (independently) mathematically proven to
 * only ever take positive values, with the specified initial value.
 */

+createPositiveWith: (double)value {
  return [self createWith: value defined: YES
	       positive: YES negative: NO constant: NO];
}

/* createPositiveWith: value min: lo max: hi -> Interval
 *
 * Create an Interval that has been (independently) mathematically proven to
 * only ever take positive values, with the specified initial range.
 */

+createPositiveWith: (double)value min: (double)lo max: (double)hi {
  return [self createWith: value min: lo max: hi defined: YES
	       positive: YES negative: NO constant: NO];
}

/* createNegative -> Interval
 *
 * Create an Interval that has been (independently) mathematically proven to
 * only ever take negative values.
 */

+createNegative {
  return [self createWith: nextafter(0.0, -1.0) defined: NO
	       positive: NO negative: YES constant: NO];
}

/* createNegativeWith: value -> Interval
 *
 * Create an Interval that has been (independently) mathematically proven to
 * only ever take negative values, with the specified initial value.
 */

+createNegativeWith: (double)value {
  return [self createWith: value defined: YES
	       positive: NO negative: YES constant: NO];
}

/* createNegativeWith: value min: lo max: hi -> Interval
 *
 * Create an Interval that has been (independently) mathematically proven to
 * only ever take negative values, with the specified initial range.
 */

+createNegativeWith: (double)value min: (double)lo max: (double)hi {
  return [self createWith: value min: lo max: hi defined: YES
	       positive: NO negative: YES constant: NO];
}

/* createConstant -> Interval
 *
 * Create an interval whose value may not change.
 */

+createConstant {
  return [self createWith: 0.0 defined: NO
	       positive: NO negative: NO constant: YES];
}

/* createConstant -> Interval
 *
 * Create an interval whose value may not change, with an initial specification
 * of that value.
 */

+createConstantWith: (double)value {
  return [self createWith: value defined: YES
	       positive: NO negative: NO constant: YES];
}

/* createConstant -> Interval
 *
 * Create an interval whose value may not change, with an initial specification
 * of the range.
 */

+createConstantWith: (double)value min: (double)lo max: (double)hi {
  return [self createWith: value min: lo max: hi defined: YES
	       positive: NO negative: NO constant: YES];
}

/* createWith: value defined: def positive: pos negative: neg constant: con
 * -> Interval (Private)
 *
 * Allocate memory for a new Interval and assign it the given value, passing
 * on whether the value is defined (i.e. has been specified by the caller
 * rather than given a default value).
 */

+createWith: (double)value defined: (BOOL)def
   positive: (BOOL)pos negative: (BOOL)neg constant: (BOOL)con {
  return [self createWith: value min: value max: value defined: def
	       positive: pos negative: neg constant: con];
}

/* createWith: value min: lo max: hi defined: def positive: pos negative: neg
 * constant: con -> Interval (Private)
 *
 * Main creation method. This passes in whether or not the initialised
 * interval has been defined, whether it is mathematically proven to be always
 * positive or always negative, and whether it is constant.
 */

+createWith: (double)value min: (double)lo max: (double)hi defined: (BOOL)def
   positive: (BOOL)pos negative: (BOOL)neg constant: (BOOL)con {
  Interval *obj = [super create: interval_zone()];

  if(lo > value || lo > hi || value > hi) {
    fprintf(stderr, "FATAL: Invalid interval initialisation: "
	    "!(%.*e <= %.*e <= %.*e)\n",
	    DBL_DIG, lo, DBL_DIG, value, DBL_DIG, hi);
    abort();
  }
  if(hi >= 0.0 && neg) {
    fprintf(stderr, "FATAL: Attempt to create a proven positive number with "
	    "negative or zero initial values: %*e, %*e, %*e\n",
	    DBL_DIG, lo, DBL_DIG, value, DBL_DIG, hi);
    abort();
  }
  if(lo <= 0.0 && pos) {
    fprintf(stderr, "FATAL: Attempt to create a proven negative number with "
	    "postive or zero initial values: %*e, %*e, %*e\n",
	    DBL_DIG, lo, DBL_DIG, value, DBL_DIG, hi);
    abort();
  }

  obj->var = value;
  obj->min = lo;
  obj->max = hi;
  obj->defined = def;
  obj->trace = NULL;
  obj->positive = pos;
  obj->negative = neg;
  obj->constant = con;

# ifdef TRACE
  if(def) {
    printf("Created %p [%g, %g, %g]\n", obj, lo, value, hi);
  }
  else {
    printf("Created %p [<undefined>]\n", obj);
  }
# endif

  nintervals++;
  return obj;
}

/* clone -> Interval
 *
 * Return a copy of self.
 */

-clone {
  return [[self class] createWith: var min: min max: max];
}

/* trace: label
 *
 * Allow this instance of Interval to be traced -- all calculations will show
 */

-trace: (const char *)label {
  trace = label;
  TRACE_SELF;
  return self;
}

/* untrace
 *
 * Switch off tracing of this Interval
 */

-untrace {
  TRACE_SELF;
  trace = NULL;
  return self;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* I/O methods */

/* bufread: (fp) -> self
 *
 * Read the values of the Interval from a file given by the file
 * pointer. If a read of the full interval fails [min, var, max], read
 * a single floating point number three times in three rounding modes:
 * RM for min, RN for var, and RP for max. This method assumes the
 * return value of fscanf is the number of successfully read
 * variables. 
 */

-bufread: (FILE *)fp {
  CHECK_CONST;
  RN;
  if(fscanf(fp, "[%lf, %lf, %lf]", &min, &var, &max) != 3) {
    long pos = 0L;

    if((pos = ftell(fp)) == -1) {
      perror("Getting file position");
      abort();
    }
    fscanf(fp, "%lf", &var);
    if(fseek(fp, pos, SEEK_SET) == -1) {
      perror("Setting file position");
      abort();
    }
    fpsetround(FP_RM);
    fscanf(fp, "%lf", &min);
    if(fseek(fp, pos, SEEK_SET) == -1) {
      perror("Setting file position");
      abort();
    }
    fpsetround(FP_RP);
    fscanf(fp, "%lf", &max);
  }
  RS;
  CHECK_POS;
  CHECK_NEG;
  defined = YES;

  TRACE_SELF;
  return self;
}

/* bufwrite -> self
 *
 * Write the interval to standard output in a simple format.
 */

-bufwrite {
  return [self bufwrite: stdout format: "%g"];
}

/* bufwrite: fp -> self
 * 
 * Write the interval to the specified file pointer with sufficient
 * accuracy to ensure that exactly the same numbers would be read back
 * in using round to nearest rounding mode.
 */

-bufwrite: (FILE *)fp {
  if(defined) {
    fprintf(fp, "[%.*e, %.*e, %.*e]",
	    minsigdig(min), min,
	    minsigdig(var), var,
	    minsigdig(max), max);
  }
  else fprintf(fp, "[<undefined>]");
  return self;
}

/* bufwrite: fp format: format -> self
 *
 * Write the interval to the specified pointer using the specified
 * format for the floating point numbers.
 */

-bufwrite: (FILE *)fp format: (char *)format {
  if(defined) {
    fprintf(fp, "[");
    fprintf(fp, format, min);
    fprintf(fp, ", ");
    fprintf(fp, format, var);
    fprintf(fp, ", ");
    fprintf(fp, format, max);
    fprintf(fp, "]");
  }
  else fprintf(fp, "[<undefined>]");
  return self;
}

/* print [POP]
 *
 * For printing an expression or a variable. Writes the interval to
 * standard output, and then drops self if it is on the stack. Use
 * this with caution -- it may be that bufwrite is a better method to
 * use, or possibly printf:
 */

-(void)print {
  [self bufwrite: stdout];
  IF_STACK_TOP(self) [pop() drop];
}

/* print: format [POP]
 *
 * As print, but lets you specify the format you want to use for each
 * floating point number. Again, use with caution.
 */

-(void)print: (char *)format {
  [self bufwrite: stdout format: format];
  IF_STACK_TOP(self) [pop() drop];
}

/* printf: format -> string [POP]
 *
 * Provides a method to call enabling the Interval to be used as part
 * of a printf statement, creating a string that can be used from
 * inside printf. For example printf("value = %g\n", num * ber) would
 * be replaced with printf("value = %s\n", [[num prod: ber] printf:
 * "%g"]). The downside of the method is that it is necessary to keep
 * a stack of strings allocated for this purpose so they can be freed
 * at a later date. Using a single class variable for this would be no
 * good, because in printf("\tnum = %s\tber = %s\n", [num printf:
 * "%g"], [ber printf: "%g"]), the arguments to printf are computed
 * first, and both would return a pointer to the same class variable,
 * meaning that either num or ber would be printed twice. Using an
 * instance variable would not work for expressions, which get
 * destroyed by this method when they are popped off the stack. Thus
 * each use of printf: should be followed by a call to printfclean so
 * that the memory allocated for the strings can be freed.
 */

-(char *)printf: (char *)format {
  char tmp[8];
  int size;
  char *printf_buffer;

  size = [self strwrite: tmp size: 8 format: format];

  printf_buffer = (char *)malloc(size + 1);
  if(printf_buffer == NULL) {
    perror("Memory allocation");
    abort();
  }
  [self strwrite: printf_buffer size: size + 1 format: format];

  push_pstack(printf_buffer);	// Save the buffer created.
  IF_STACK_TOP(self) [pop() drop];
  return printf_buffer;
}

-(char *)printfVar: (char *)format {
  char tmp[8];
  int size;
  char *printf_buffer;

  size = [self strwriteVar: tmp size: 8 format: format];

  printf_buffer = (char *)malloc(size + 1);
  if(printf_buffer == NULL) {
    perror("Memory allocation");
    abort();
  }
  [self strwriteVar: printf_buffer size: size + 1 format: format];

  push_pstack(printf_buffer);	// Save the buffer created.
  IF_STACK_TOP(self) [pop() drop];
  return printf_buffer;
}

-(char *)printfRange: (char *)format {
  char tmp[8];
  int size;
  char *printf_buffer;

  size = [self strwriteRange: tmp size: 8 format: format];

  printf_buffer = (char *)malloc(size + 1);
  if(printf_buffer == NULL) {
    perror("Memory allocation");
    abort();
  }
  [self strwriteRange: printf_buffer size: size + 1 format: format];

  push_pstack(printf_buffer);	// Save the buffer created.
  IF_STACK_TOP(self) [pop() drop];
  return printf_buffer;
}

/* printfclean
 *
 * As explained in the printf: method, it is necessary to build a
 * stack of string pointers containing the buffers created to store
 * the Interval strings used as part of printf function calls. This
 * method cleans up the stack, freeing all the allocated strings. It
 * should follow any printf function call.
 */

+(void)printfclean {
  clear_pstack();
}

/* strread: buffer -> self
 *
 * Attempt to read from the buffer passed in as argument an
 * interval. Just as for bufread, if the buffer does not contain an
 * Interval format [min, var, max], then read a single floating point
 * number three times in each appropriate rounding mode. 
 */

-strread: (char *)buffer {
  CHECK_CONST;
  RN;
  if(sscanf(buffer, "[%lf, %lf, %lf]", &min, &var, &max) != 3) {
    sscanf(buffer, "%lf", &var);
    fpsetround(FP_RM);
    sscanf(buffer, "%lf", &min);
    fpsetround(FP_RP);
    sscanf(buffer, "%lf", &max);
  }
  RS;
  defined = YES;
  CHECK_POS;
  CHECK_NEG;

  TRACE_SELF;
  return self;
}

/* strwrite: buffer size: bufsz -> number of characters formatted
 *
 * Like bufwrite:, but to a string of specified size instead of to a
 * file. Returns the number of characters required to store the
 * Interval string, not including the NULL terminator. Relies on
 * snprintf returning this number of characters.
 */

-(int)strwrite: (char *)buffer size: (size_t)bufsz {
  if(defined) {
    return snprintf(buffer, bufsz, "[%.*e, %.*e, %.*e]",
		    minsigdig(min), min,
		    minsigdig(var), var,
		    minsigdig(max), max);
  }
  else return snprintf(buffer, bufsz, "[<undefined>]");
}

-(int)strwriteVar: (char *)buffer size: (size_t)bufsz {
  if(defined) {
    return snprintf(buffer, bufsz, "%.*e", minsigdig(var), var);
  }
  else return snprintf(buffer, bufsz, "<undefined>");
}

-(int)strwriteRange: (char *)buffer size: (size_t)bufsz {
  if(defined) {
    return snprintf(buffer, bufsz, "[%.*e, %.*e]",
		    minsigdig(min), min,
		    minsigdig(max), max);
  }
  else {
    return snprintf(buffer, bufsz, "<undefined>");
  }
}

/* strwrite: buffer size: bufsz format: format -> number of chars formatted
 *
 * Like strwrite:size:, but allows the format for the floating point
 * numbers to be specified.
 */

-(int)strwrite: (char *)buffer size: (size_t)bufsz format: (char *)format {
  char *tmp_format = (char *)malloc((strlen(format) * 3) + 7);
  int ret_val;

  if(tmp_format == NULL) {
    perror("Memory allocation");
    abort();
  }
  if(defined) {
    sprintf(tmp_format, "[%s, %s, %s]", format, format, format);
    ret_val = snprintf(buffer, bufsz, tmp_format, min, var, max);
  }
  else ret_val = snprintf(buffer, bufsz, "[<undefined>]");
  free(tmp_format);
  return ret_val;
}

-(int)strwriteVar: (char *)buffer size: (size_t)bufsz format: (char *)format {
  if(defined) {
    return snprintf(buffer, bufsz, format, var);
  }
  else return snprintf(buffer, bufsz, "<undefined>");
}

-(int)strwriteRange: (char *)buffer size: (size_t)bufsz format: (char *)format
{
  char *tmp_format = (char *)malloc((strlen(format) * 2) + 5);
  int ret_val;

  if(tmp_format == NULL) {
    perror("Memory allocation");
    abort();
  }
  if(defined) {
    sprintf(tmp_format, "[%s, %s]", format, format);
    ret_val = snprintf(buffer, bufsz, tmp_format, min, max);
  }
  else {
    ret_val = snprintf(buffer, bufsz, "[<undefined>]");
  }
  free(tmp_format);
  return ret_val;
}


/* read: fildes -> self
 *
 * Attempt to read the interval from a binary file with file
 * descriptor fildes.
 */

-read: (int)fildes {
  CHECK_CONST;
  if(read(fildes, &min, sizeof(double)) == -1) {
    perror("Interval: file read error");
    abort();
  }
  if(read(fildes, &var, sizeof(double)) == -1) {
    perror("Interval: file read error");
    abort();
  }
  if(read(fildes, &max, sizeof(double)) == -1) {
    perror("Interval: file read error");
    abort();
  }
  CHECK_POS;
  CHECK_NEG;
  defined = YES;

  TRACE_SELF;
  return self;
}

/* write: fildes -> self
 *
 * Attempt to write the interval to a binary file with file descriptor
 * fildes.
 */

-write: (int)fildes {
  if(!defined) {
    if(!hide_warn) {
      fprintf(stderr, "WARNING: Writing undefined interval\n");
    }
  }
  if(write(fildes, &min, sizeof(double)) == -1) {
    perror("Interval: file write error");
    abort();
  }
  if(write(fildes, &var, sizeof(double)) == -1) {
    perror("Interval: file write error");
    abort();
  }
  if(write(fildes, &max, sizeof(double)) == -1) {
    perror("Interval: file write error");
    abort();
  }
  return self;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Accessor methods */

/* set: value -> self [POP]
 *
 * Set self's Interval parameters to be the same as those of value. If
 * value is on the stack, drop it.
 */

-set: (Interval *)value {
  CHECK_CONST;

  if(!value->defined) {
    warn_defined(value, _cmd);
  }

  [value get: &var min: &min max: &max];
  defined = YES;

  CHECK_POS;
  CHECK_NEG;

  TRACE_SELF_ARG(value);

  IF_STACK_TOP(value) [pop() drop];

  return self;
}

/* get: varptr min: minptr max: maxptr -> self
 *
 * Allow access to all the interval parameters.
 */

-get: (double *)varptr min: (double *)minptr max: (double *)maxptr {
  if(varptr != NULL) (*varptr) = var;
  if(minptr != NULL) (*minptr) = min;
  if(maxptr != NULL) (*maxptr) = max;
  return self;
}

/* getVar
 *
 * Return the var parameter.
 */

-(double)getVar {
  CHECK_DEF;
  return var;
}

/* getMin
 *
 * Return the min parameter.
 */

-(double)getMin {
  CHECK_DEF;
  return min;
}

/* getMax
 *
 * Return the max parameter.
 */

-(double)getMax {
  CHECK_DEF;
  return max;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Arithmetic operator methods */

/* add: value -> self [POP]
 *
 * Implement += operator. For addition, it is safe to assume that
 * [minA, maxA] + [minB, maxB] = [minA + minB, maxA + maxB].
 */

-add: (Interval *)value {
  double vvar, vmin, vmax;

  CHECK_DEF_ARG(value);
  CHECK_CONST;

  [value get: &vvar min: &vmin max: &vmax];

# ifdef TRACE
  printf("[%g, %g, %g] + [%g, %g, %g] = ", min, var, max, vmin, vvar, vmax);
# endif

  RM;
  min += vmin;
  fpsetround(FP_RP);
  max += vmax;
  fpsetround(FP_RN);
  var += vvar;
  RS;

  CHECK_POS;
  CHECK_NEG;

# ifdef TRACE
  printf("[%g, %g, %g]\n", min, var, max);
# endif

  TRACE_SELF_ARG(value);

  IF_STACK_TOP(value) [pop() drop];
  
  return self;
}

/* sum: value -> Interval [PUSH, POP]
 *
 * Implement operator +. This method calls += with a possibly newly
 * created Interval object that is pushed on the stack. The only
 * circumstances in which a new object is not created and put on the
 * stack is when self is already on the stack. However, if value is
 * also on the stack, it will be on top of self, so we need to take
 * value off first, if required.
 *
 * <private>Would this method name be better named plus? That way all
 * x= operators have three letter method names and all x operators
 * have four letter names.</private>
 */

-sum: (Interval *)value {
  Interval *value_if_on_stack = nil;
  Interval *operand = nil;
  id retval;
  
  IF_STACK_TOP(value) value_if_on_stack = pop();
  operand = STACK_TOP(self) ? self : [self clone];

  retval = [operand add: value];

  if(operand != self) push(operand);
  if(value_if_on_stack != nil) [value_if_on_stack drop];

  return retval;
}

/* inc -> self
 *
 * Implement prefix operator ++.
 */

-inc {
# ifdef TRACE
  printf("[%g, %g, %g]++ = ", min, var, max);
# endif

  CHECK_DEF;
  CHECK_CONST;

  RM;
  min += 1.0;
  fpsetround(FP_RP);
  max += 1.0;
  fpsetround(FP_RN);
  var += 1.0;
  RS;

  CHECK_POS;
  CHECK_NEG;

  TRACE_SELF;

# ifdef TRACE
  printf("[%g, %g, %g]\n", min, var, max);
# endif

  return self;
}

/* sub: value -> self [POP]
 *
 * Implement operator -=. Here it is safe to assume that [minA, maxA]
 * - [minB, maxB] = [minA - maxB, maxA - minB].
 */

-sub: (Interval *)value {
  double vvar, vmin, vmax;

  CHECK_DEF_ARG(value);
  CHECK_CONST;

  [value get: &vvar min: &vmin max: &vmax];

# ifdef TRACE
  printf("[%g, %g, %g] - [%g, %g, %g] = ", min, var, max, vmin, vvar, vmax);
# endif

  RM;
  min -= vmax;
  fpsetround(FP_RP);
  max -= vmin;
  fpsetround(FP_RN);
  var -= vvar;
  RS;

  CHECK_POS;
  CHECK_NEG;

# ifdef TRACE
  printf("[%g, %g, %g]\n", min, var, max);
# endif

  TRACE_SELF_ARG(value);

  IF_STACK_TOP(value) [pop() drop];
  
  return self;
}

/* diff: value -> Interval [PUSH, POP]
 *
 * Implement binary operator -. Just as for sum:, the result of this
 * method is always left on the stack.
 */

-diff: (Interval *)value {
  Interval *value_if_on_stack = nil;
  Interval *operand = nil;
  id retval;
  
  IF_STACK_TOP(value) value_if_on_stack = pop();
  operand = STACK_TOP(self) ? self : [self clone];

  retval = [operand sub: value];

  if(operand != self) push(operand);
  if(value_if_on_stack != nil) [value_if_on_stack drop];

  return retval;
}

/* dec -> self
 *
 * Implement prefix operator --.
 */

-dec {
# ifdef TRACE
  printf("[%g, %g, %g]-- =", min, var, max);
# endif

  CHECK_DEF;
  CHECK_CONST;

  RM;
  min -= 1.0;
  fpsetround(FP_RP);
  max -= 1.0;
  fpsetround(FP_RN);
  var -= 1.0;
  RS;

  CHECK_POS;
  CHECK_NEG;

  TRACE_SELF;

# ifdef TRACE
  printf("[%g, %g, %g]\n", min, var, max);
# endif

  return self;
}

/* mul: value -> self [POP]
 *
 * Implement operator *=. This uses the BINARY_OP macro, which checks
 * all combinations of the multiplicands' interval extremities to find
 * the minimum and maximum result. 
 */

-mul: (Interval *)value {
  double vmin, vmax, vvar;

  CHECK_DEF_ARG(value);
  CHECK_CONST;

  [value get: &vvar min: &vmin max: &vmax];

# ifdef TRACE
  printf("[%g, %g, %g] * [%g, %g, %g] = ", min, var, max, vmin, vvar, vmax);
# endif

  BINARY_OP(*, min, max, var, min, max, var, vmin, vmax, vvar);

  CHECK_POS;
  CHECK_NEG;

# ifdef TRACE
  printf("[%g, %g, %g]\n", min, var, max);
# endif

  IF_STACK_TOP(value) [pop() drop];
  
  return self;
}

/* prod: value -> Interval [PUSH, POP]
 *
 * Implement binary operator *.
 */

-prod: (Interval *)value {
  Interval *value_if_on_stack = nil;
  Interval *operand = nil;
  id retval;
  
  IF_STACK_TOP(value) value_if_on_stack = pop();
  operand = STACK_TOP(self) ? self : [self clone];

  retval = [operand mul: value];

  if(operand != self) push(operand);
  if(value_if_on_stack != nil) [value_if_on_stack drop];

  return retval;
}

/* div: value -> self [POP]
 *
 * Implement operator /=. Just as for mul:, all combinations of
 * minimum and maximum values of the operands have to be
 * inspected. However, with div: there is the additional complexity of
 * what to do when the divisor has a range that includes
 * zero. Officially, the operation is undefined in such circumstances
 * (Alefeld & Herzberger, 1983, p. 2), but in IEEE arithmetic,
 * anything (except zero) divided by zero is the correctly signed
 * infinity (Section 7.2). If the divisor contains zero, and the
 * dividend does not, then the result has minimum -Inf and maximum
 * +Inf, regardless of its sign. This is because anything that
 * contains 0.0 also contains -0.0 since 0.0 == -0.0 (Section 5.7),
 * and the IEEE standard defines the result of division to be the XOR
 * of the signs of its operands (Section 6.3). Hence for all +ve x, x
 * / 0.0 is +Inf and x / -0.0 is -Inf, whilst -x / 0.0 is -Inf and -x
 * / -0.0 is +Inf. In section 7.1, the IEEE standard states that
 * 0.0/0.0 is an invalid operation. Thus if the dividend spans zero,
 * the result is potentially NaN, and it seems sensible to set both
 * min and max to this value. 
 */

-div: (Interval *)value {
  double vmin, vmax, vvar;

  CHECK_DEF_ARG(value);
  CHECK_CONST;

  [value get: &vvar min: &vmin max: &vmax];

# ifdef TRACE
  printf("[%g, %g, %g] / [%g, %g, %g] = ", min, var, max, vmin, vvar, vmax);
# endif

  if((vmin < 0.0 && vmax > 0.0)
     || (vmin < 0.0 && fpclass(vmax) == FP_PZERO)
     || (fpclass(vmin) == FP_NZERO && vmax > 0.0)
     || (fpclass(vmin) == FP_NZERO && fpclass(vmax) == FP_PZERO)) {
				// The divisor contains +/-zero.
    if(strict_div) {
      if(min <= 0.0 && max >= 0.0) {
				// The dividend contains zero
	min = max = 0.0 / 0.0;	// This is defined as NaN by IEEE 754
      }
      else {			// The dividend does not contain zero
	min = -1.0 / 0.0;		// Must be -Inf by IEEE 754
	max = 1.0 / 0.0;		// Must be +Inf by IEEE 754
      }
    }
    else {
      if(!hide_warn) {
	fprintf(stderr, "WARNING: Division by range containing zero: ");
	[value bufwrite: stderr];
	fprintf(stderr, "\n");
      }
      if(vvar < 0.0) {
	vmax = nextafter(vvar, vmax);
      }
      else if(vvar > 0.0) {
	vmin = nextafter(vvar, vmin);
      }
      BINARY_OP(/, min, max, var, min, max, var, vmin, vmax, vvar);
    }
    if(nan_crash) {
      fprintf(stderr, "FATAL: Division by range containing zero: ");
      [value bufwrite: stderr];
      fprintf(stderr, "\n");
      abort();
    }
    var /= vvar;
  }
  else {			// The divisor does not contain +/-zero
    BINARY_OP(/, min, max, var, min, max, var, vmin, vmax, vvar);
  }

  CHECK_POS;
  CHECK_NEG;

  TRACE_SELF_ARG(value);

# ifdef TRACE
  printf("[%g, %g, %g]\n", min, var, max);
# endif

  IF_STACK_TOP(value) [pop() drop];
  
  return self;
}

/* quot: value -> Interval [PUSH, POP]
 *
 * Implement operator /.
 */

-quot: (Interval *)value {
  Interval *value_if_on_stack = nil;
  Interval *operand = nil;
  id retval;
  
  IF_STACK_TOP(value) value_if_on_stack = pop();
  operand = STACK_TOP(self) ? self : [self clone];

  retval = [operand div: value];

  if(operand != self) push(operand);
  if(value_if_on_stack != nil) [value_if_on_stack drop];

  return retval;
}

/* neg -> Interval [PUSH]
 *
 * Implement unary operator -. This puts a negated value of self onto
 * the stack (unless self is already on top of the stack).
 */

-neg {
  double vmin, vmax, vvar;

  CHECK_DEF;

  RM;
  vmin = -max;
  fpsetround(FP_RP);
  vmax = -min;
  fpsetround(FP_RN);
  vvar = -var;
  RS;

# ifdef TRACE
  printf("-[%g, %g, %g] = [%g, %g, %g]\n", min, var, max, vmin, vvar, vmax);
# endif

  IF_STACK_TOP(self) {
    max = vmax;
    min = vmin;
    var = vvar;
    return self;
  }
  else return push([[self class] createWith: vvar min: vmin max: vmax]);
}

/* integer -> int [POP]
 *
 * Implement a cast to int. This issues a warning if a different integer is
 * generated from max and min. If self is on top of the stack it will be
 * popped.
 */

-(int)integer {
  int imax, imin, ivar;

  //  if(max > (double)(MAXINT) || min < (double)(-(MAXINT) - 1)) {
  /* gcc doesn't like MAXINT in values.h suggesting parentheses around
   * the shift, and values.h doesn't appear in Cygwin, which
   * necessitates the following:
   */
  int minint = (1 << ((sizeof(int) * 8) - 1));
  int maxint = ~minint;
  if(max > (double)maxint || min < (double)minint) {
    if(!hide_warn) {
      fprintf(stderr, "WARNING: Overflow converting ");
      [self bufwrite: stderr];
      fprintf(stderr, " to an integer\n");
    }
  }

  CHECK_DEF;

  RP;
  imax = (max < 0.0) ? (int)ceil(max) : (int)floor(max);
  fpsetround(FP_RM);
  imin = (min < 0.0) ? (int)ceil(min) : (int)floor(min);
  fpsetround(FP_RN);
  ivar = (var < 0.0) ? (int)ceil(var) : (int)floor(var);
  RS;

  if(imax != imin) {
    if(!hide_warn) {
      fprintf(stderr, "WARNING: uncertainty converting ");
      [self bufwrite: stderr];
      fprintf(stderr, "to an integer: it could be any integer between %d and "
	      "%d inclusive\n", imin, imax);
    }
  }

  IF_STACK_TOP(self) [pop() drop];

  return ivar;
}

/* rint -> Interval [PUSH]
 *
 * Implement round to nearest integer. If self is not on top of the stack,
 * then push the rounded value. If min <= var <= max, then we can be sure
 * that rint(min) <= rint(var) <= rint(max).
 */

-(Interval *)rint {
  double vmin, vmax, vvar;

  CHECK_DEF;

  RM;
  vmin = rint(min);
  fpsetround(FP_RP);
  vmax = rint(max);
  fpsetround(FP_RN);
  vvar = rint(var);
  RS;

# ifdef TRACE
  printf("rint([%g, %g, %g]) = [%g, %g, %g]\n",
	 min, var, max, vmin, vvar, vmax);
# endif

  IF_STACK_TOP(self) {
    max = vmax;
    min = vmin;
    var = vvar;
    return self;
  }
  else return push([[self class] createWith: vvar min: vmin max: vmax]);
}

/* fabs -> Interval [PUSH]
 *
 * Return the absolute value of self. If self is not on top of the stack,
 * then push the absolute value.
 */

-(Interval *)fabs {
  double vmin = min, vmax = max, vvar = var;

  CHECK_DEF;

  if(min < 0.0 && max < 0.0) {
    vmin = -max;
    vmax = -min;
    vvar = -var;
  }
  else if(min < 0.0 && max > 0.0) {
    vmin = 0.0;
    vmax = (-min > max ? -min : max);
    vvar = (vvar < 0.0 ? -vvar : vvar);
  }
  else if(min > 0.0 && max > 0.0) {
    vmin = min;
    vmax = max;
    vvar = var;
  }
  if(fpclass(min) == FP_NZERO) vmin = 0.0;
  if(fpclass(max) == FP_NZERO) vmax = 0.0;
  if(fpclass(var) == FP_NZERO) vvar = 0.0;

# ifdef TRACE
  printf("fabs([%g, %g, %g]) = [%g, %g, %g]\n",
	 min, var, max, vmin, vvar, vmax);
# endif

  IF_STACK_TOP(self) {
    max = vmax;
    min = vmin;
    var = vvar;
    return self;
  }
  else return push([[self class] createWith: vvar min: vmin max: vmax]);
}
  
/* floor -> Interval [PUSH]
 *
 * Push a new interval given as the range of values of the floor of
 * the current interval. For all x, floor(x) <= x, so if a <= b <= c
 * then floor(a) <= floor(b) <= floor(c).
 */

-(Interval *)floor {
  double vmin, vmax, vvar;

  CHECK_DEF;

  RM;
  vmin = floor(min);
  fpsetround(FP_RP);
  vmax = floor(max);
  fpsetround(FP_RN);
  vvar = floor(var);
  RS;

# ifdef TRACE
  printf("floor([%g, %g, %g]) = [%g, %g, %g]\n",
	 min, var, max, vmin, vvar, vmax);
# endif

  IF_STACK_TOP(self) {
    max = vmax;
    min = vmin;
    var = vvar;
    return self;
  }
  else return push([[self class] createWith: vvar min: vmin max: vmax]);
}

/* ceil -> Interval [PUSH]
 *
 * Push a new interval given as the range of values for the ceiling of
 * the current interval. For all x, ceil(x) >= x, so if a <= b <= c
 * then ceil(a) <= ceil(b) <= ceil(c).
 */

-(Interval *)ceil {
  double vmin, vmax, vvar;

  CHECK_DEF;

  RM;
  vmin = ceil(min);
  fpsetround(FP_RP);
  vmax = ceil(max);
  fpsetround(FP_RN);
  vvar = ceil(var);
  RS;

# ifdef TRACE
  printf("ceil([%g, %g, %g]) = [%g, %g, %g]\n",
	 min, var, max, vmin, vvar, vmax);
# endif

  IF_STACK_TOP(self) {
    max = vmax;
    min = vmin;
    var = vvar;
    return self;
  }
  else return push([[self class] createWith: vvar min: vmin max: vmax]);
}

/* flatten
 *
 * Flatten the interval, setting min = max = var. Issue a warning message
 * that this has been done if min != var or max != var.
 */

-(void)flatten {
# ifdef TRACE
  double vmin = min, vvar = var, vmax = max;
# endif

  CHECK_DEF;

  if(min != var || max != var) {
    if(!hide_warn) {
      fprintf(stderr, "WARNING: flattened [%g, %g, %g]\n", min, var, max);
    }
  }

  min = max = var;

# ifdef TRACE
  printf("flatten([%g, %g, %g]) = [%g, %g, %g]\n", vmin, vvar, vmax,
	 min, max, var);
# endif

  TRACE_SELF;
}

/* constrainMin: -> changed? [POP]
 *
 * Lower bound self by the interval passed in as argument. This method
 * is used to replace code that says if(x < z) x = z; If z is an
 * interval, there is an issue about the nature of the bounds. This
 * method is designed to say, "set x such that there is no member of z
 * such that there exists a member of x such that x < z". Potentially
 * this can result in a change to var.
 *
 * The method returns a Boolean to indicate whether or not self had to
 * be changed to perform the constraint. This allows other actions to
 * be coded for if a constraint took place. 
 */

-(BOOL)constrainMin: (Interval *)lo {
  double vmin, vmax, vvar;
  BOOL changed = NO;

  CHECK_DEF;
  CHECK_CONST;

  CHECK_DEF_ARG(lo);

  [lo get: &vvar min: &vmin max: &vmax];

# ifdef TRACE
  printf("constrain([%g, %g, %g] >= [%g, %g, %g])",
	 min, var, max, vmin, vvar, vmax);
# endif

  if(var < vvar) {
    var = vvar;
    changed = YES;
  }
  if(min < vmax) {
    min = vmax;
    if(var < min) var = min;	// N.B. var changed.
    changed = YES;
  }
  if(max < vmax) {
    max = vmax;
    changed = YES;
  }

  CHECK_POS;
  CHECK_NEG;

# ifdef TRACE
  printf(" = [%g, %g, %g]\n", min, var, max);
# endif

  TRACE_SELF_ARG(lo);

  IF_STACK_TOP(lo) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return changed;
}

/* constrainMax: -> Boolean [POP]
 *
 * Upper bound self by the interval passed in as argument. This method
 * is used to replace code that says if(x > y) x = y; If y is an
 * interval, there is an issue about the nature of the bounds. This
 * method is designed to say, "set x such that there is no member of y
 * such that there exists a member of x such that x > y". Potentially
 * this can result in a change to var.
 *
 * The method returns a Boolean to indicate whether or not self had to
 * be changed to perform the constraint. This allows other actions to
 * be coded for if a constraint took place. 
 */

-(BOOL)constrainMax: (Interval *)hi {
  double vmin, vmax, vvar;
  BOOL changed = NO;

  CHECK_DEF;
  CHECK_CONST;

  CHECK_DEF_ARG(hi);

  [hi get: &vvar min: &vmin max: &vmax];

# ifdef TRACE
  printf("constrain([%g, %g, %g] <= [%g, %g, %g])",
	 min, var, max, vmin, vvar, vmax);
# endif

  if(var > vvar) {
    var = vvar;
    changed = YES;
  }
  if(min > vmin) {
    min = vmin;
    changed = YES;
  }
  if(max > vmin) {
    max = vmin;
    if(var > max) var = max;	// N.B. var changed.
    changed = YES;
  }
  
  CHECK_POS;
  CHECK_NEG;

# ifdef TRACE
  printf(" = [%g, %g, %g]\n", min, var, max);
# endif

  TRACE_SELF_ARG(hi);

  IF_STACK_TOP(hi) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return changed;
}

/* constrainPlusMinus: -> self [POP]
 *
 * Bound self by plus/minus the argument. This method is used to
 * replace code that says if(x > y) x = y; else if(x < -y) x = -y; If
 * y is an interval, there is an issue about the bounds. This method
 * is designed to say, "set x such that there is no member of y such
 * that there exists a member of x such that x < -y or x >
 * y". Potentially this can result in a change to var. If the argument
 * spans zero then this is an error.
 *
 * The method returns a Boolean to indicate whether or not self had to
 * be changed to perform the constraint. This allows other actions to
 * be coded for if a constraint took place. 
 */

-(BOOL)constrainPlusMinus: (Interval *)pm {
  double vmin, vmax, vvar;
  BOOL changed = NO;
  Interval *abspm;

  CHECK_DEF;
  CHECK_CONST;

  CHECK_DEF_ARG(pm);

  [pm get: &vvar min: &vmin max: &vmax];

  if(vmin < 0.0 && vmax > 0.0) {
    fprintf(stderr, "FATAL: contrain to +/- [%g, %g, %g]\n", vmin, vvar, vmax);
    abort();
  }

  abspm = [[Interval create] set: [pm fabs]];

  [abspm get: &vvar min: &vmin max: &vmax];

# ifdef TRACE
  printf("constrain(|[%g, %g, %g]| <= [%g, %g, %g])",
	 min, var, max, vmin, vvar, vmax);
# endif

  /* Deliberate use of -vmin rather than -vmax, the latter looking more
   * like constrainMin:, as -vmin is the maximum of the lower bound.
   */

  if(var < -vvar) {
    var = -vvar;
    changed = YES;
  }
  if(min < -vmin) {
    min = -vmin;
    if(var < min) var = min;	// N.B. var changed.
    changed = YES;
  }
  if(max < -vmin) {
    max = -vmin;
    changed = YES;
  }

  if(var > vvar) {
    var = vvar;
    changed = YES;
  }
  if(min > vmin) {
    min = vmin;
    changed = YES;
  }
  if(max > vmin) {
    max = vmin;
    if(var > max) var = max;	// N.B. var changed.
    changed = YES;
  }
  
  CHECK_POS;
  CHECK_NEG;

# ifdef TRACE
  printf(" = [%g, %g, %g]\n", min, var, max);
# endif

  TRACE_SELF_ARG(pm);

  IF_STACK_TOP(pm) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];
  [abspm drop];

  return changed;
}

/* exp -> Interval [PUSH]
 *
 * Push a new interval given as the range of values for the base of natural
 * logarithms raised to the power of the current interval. Since e^x is a 
 * monotone increasing function of x, if min <= var <= max then e^min <=
 * e^var <= e^max.
 */

-(Interval *)exp {
  double vmin, vmax, vvar;

  CHECK_DEF;

  RM;
  vmin = exp(min);
  fpsetround(FP_RP);
  vmax = exp(max);
  fpsetround(FP_RN);
  vvar = exp(var);
  RS;

# ifdef TRACE
  printf("exp([%g, %g, %g]) = [%g, %g, %g]\n",
	 min, var, max, vmin, vvar, vmax);
# endif

  IF_STACK_TOP(self) {
    min = vmin;
    max = vmax;
    var = vvar;
    return self;
  }
  else return push([[self class] createWith: vvar min: vmin max: vmax]);
}

/* expm1 -> Interval [PUSH]
 *
 * Push a new interval given as the range of values for (the base of
 * natural logarithms raised to the power of the current interval)
 * minus 1. Since e^x - 1 is a monotone increasing function of x, if min
 * <= var <= max then e^min - 1 <= e^var - 1 <= e^max - 1. 
 */

-(Interval *)expm1 {
  double vmin, vmax, vvar;

  CHECK_DEF;

  RM;
  vmin = expm1(min);
  fpsetround(FP_RP);
  vmax = expm1(max);
  fpsetround(FP_RN);
  vvar = expm1(var);
  RS;

# ifdef TRACE
  printf("expm1([%g, %g, %g]) = [%g, %g, %g]\n",
	 min, var, max, vmin, vvar, vmax);
# endif

  IF_STACK_TOP(self) {
    min = vmin;
    max = vmax;
    var = vvar;
    return self;
  }
  else return push([[self class] createWith: vvar min: vmin max: vmax]);
}

/* sqrt -> Interval [PUSH]
 *
 * Push a new interval given as the range of values for the square root of
 * the current interval. Since sqrt(x) is a monotone increasing function of x,
 * then if min <= var <= max, sqrt(min) <= sqrt(var) <= sqrt(max). The only
 * exception is if min is negative and max non-negative. In this case a warning
 * is issued.
 */

-(Interval *)sqrt {
  double vmin, vmax, vvar;

  CHECK_DEF;

  if(min < 0.0) {
    if(nan_crash) {
      fprintf(stderr, "FATAL: Square root of an interval "
	      "with negative elements: ");
      [self bufwrite: stderr];
      fprintf(stderr, "\n");
      abort();
    }
    else if(!hide_warn) {
      fprintf(stderr, "WARNING: Square root of an interval with negative"
	      " elements: ");
      [self bufwrite: stderr];
      fprintf(stderr, "\n");
    }
  }

  RM;
  vmin = sqrt(min);
  fpsetround(FP_RP);
  vmax = sqrt(max);
  fpsetround(FP_RN);
  vvar = sqrt(var);
  RS;

# ifdef TRACE
  printf("sqrt([%g, %g, %g]) = [%g, %g, %g]\n",
	 min, var, max, vmin, vvar, vmax);
# endif

  IF_STACK_TOP(self) {
    min = vmin;
    max = vmax;
    var = vvar;
    return self;
  }
  else return push([[self class] createWith: vvar min: vmin max: vmax]);
}

/* squared -> Interval [PUSH]
 *
 * Push a new interval onto the stack which is the square of self. The result
 * of this operation is guaranteed to be positive or zero. If the interval
 * spans zero, therefore, the minimum is zero.
 */

-(Interval *)squared {
  double vmin, vmax, vvar;

  CHECK_DEF;

  if(min < 0.0 && max > 0.0) {
    double minsq, maxsq;

    RP;
    minsq = min * min;
    maxsq = max * max;
    RS;

    vmin = 0.0;
    vmax = (minsq > maxsq) ? minsq : maxsq;
  }
  else {
    if(min > 0.0) {
      RM;
      vmin = min * min;
      fpsetround(FP_RP);
      vmax = max * max;
      RS;
    }
    else {
      RM;
      vmin = max * max;
      fpsetround(FP_RP);
      vmax = min * min;
      RS;
    }
  }

  RN;
  vvar = var * var;
  RS;


# ifdef TRACE
  printf("[%g, %g, %g]^2 = [%g, %g, %g]\n",
	 min, var, max, vmin, vvar, vmax);
# endif

  IF_STACK_TOP(self) {
    min = vmin;
    max = vmax;
    var = vvar;
    return self;
  }
  else return push([[self class] createWith: vvar min: vmin max: vmax]);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Comparison operator methods */

/* eq: value -> Boolean [POP]
 *
 * Tests for equality between the var component of self and value, and
 * checks for consistency with Intervals, issuing a warning in the
 * event that equality is true for var, but possibly false with the
 * intervals, or equality is false for var, but possibly true with
 * intervals.
 */

-(BOOL)eq: (Interval *)value {
  double vmin, vmax, vvar;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: &vvar min: &vmin max: &vmax];

  if(var == vvar) {
    if(possibly_ne(min, max, vmin, vmax))
      warn_uncertainty(self, value, "eq but Mne");
    retval = YES;
  }
  else {
    if(possibly_eq(min, max, vmin, vmax))
      warn_uncertainty(self, value, "!eq but Meq");
    retval = NO;
  }

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}

/* ne: value -> Boolean [POP]
 *
 * Tests for inequality between the var components of self and
 * value. As for eq:, checks consistency with the intervals, and
 * issues a warning.
 */

-(BOOL)ne: (Interval *)value {
  double vmin, vmax, vvar;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: &vvar min: &vmin max: &vmax];

  if(var != vvar) {
    if(possibly_eq(min, max, vmin, vmax))
      warn_uncertainty(self, value, "ne but Meq");
    retval = YES;
  }
  else {
    if(possibly_ne(min, max, vmin, vmax))
      warn_uncertainty(self, value, "!ne but Mne");
    retval = NO;
  }

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;  
}

/* gt: value -> Boolean [POP]
 *
 * Tests if self is greater than value using the var component,
 * checking for consistency with the intervals.
 */

-(BOOL)gt: (Interval *)value {
  double vmin, vmax, vvar;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: &vvar min: &vmin max: &vmax];

  if(var > vvar) {
    if(possibly_le(min, max, vmin, vmax))
      warn_uncertainty(self, value, "gt but Mle");
    retval = YES;
  }
  else {
    if(possibly_gt(min, max, vmin, vmax))
      warn_uncertainty(self, value, "!gt but Mgt");
    retval = NO;
  }

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;  
}

/* le: value -> Boolean [POP]
 *
 * Tests if self is less than or equal to value using the var
 * component, checking for consistency with the intervals.
 */

-(BOOL)le: (Interval *)value {
  double vmin, vmax, vvar;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: &vvar min: &vmin max: &vmax];

  if(var <= vvar) {
    if(possibly_gt(min, max, vmin, vmax))
      warn_uncertainty(self, value, "le but Mgt");
    retval = YES;
  }
  else {
    if(possibly_le(min, max, vmin, vmax))
      warn_uncertainty(self, value, "!le but Mle");
    retval = NO;
  }

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;  
}

/* lt: value -> Boolean [POP]
 *
 * Tests if self is less than value using the var component, checking
 * for consistency with the intervals.
 */

-(BOOL)lt: (Interval *)value {
  double vmin, vmax, vvar;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: &vvar min: &vmin max: &vmax];

  if(var < vvar) {
    if(possibly_ge(min, max, vmin, vmax))
      warn_uncertainty(self, value, "lt but Mge");
    retval = YES;
  }
  else {
    if(possibly_lt(min, max, vmin, vmax))
      warn_uncertainty(self, value, "!lt but Mlt");
    retval = NO;
  }

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;  
}

/* ge: value -> Boolean [POP]
 *
 * Tests if self is greater than or equal to value using the var
 * component, checking for consistency with the intervals.
 */

-(BOOL)ge: (Interval *)value {
  double vmin, vmax, vvar;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: &vvar min: &vmin max: &vmax];

  if(var >= vvar) {
    if(possibly_lt(min, max, vmin, vmax))
      warn_uncertainty(self, value, "ge but Mlt");
    retval = YES;
  }
  else {
    if(possibly_ge(min, max, vmin, vmax))
      warn_uncertainty(self, value, "!ge but Mge");
    retval = NO;
  }

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;  
}

/* Meq: value -> Boolean [POP]
 *
 * Returns true if self is possibly equal to value, that is, if there
 * exists a member of self's range of values that is equal to a member
 * of value's. Returns false otherwise.
 */

-(BOOL)Meq: (Interval *)value {
  double vmin, vmax;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: NULL min: &vmin max: &vmax];

  retval = possibly_eq(min, max, vmin, vmax);

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}

/* Mne: value -> Boolean [POP]
 *
 * Returns true iff there exists a member of [self->min, self->max]
 * that is not equal to a member of [value->min, value->max].
 */

-(BOOL)Mne: (Interval *)value {
  double vmin, vmax;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: NULL min: &vmin max: &vmax];

  retval = possibly_ne(min, max, vmin, vmax);

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}

/* Mgt: value -> Boolean [POP]
 *
 * Returns true iff there exists a member of [self->min, self->max]
 * that is greater than a member of [value->min, value->max].
 */

-(BOOL)Mgt: (Interval *)value {
  double vmin, vmax;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: NULL min: &vmin max: &vmax];

  retval = possibly_gt(min, max, vmin, vmax);

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}

/* Mle: value -> Boolean [POP]
 *
 * Returns true iff there exists a member of [self->min, self->max]
 * that is less than or equal to a member of [value->min, value->max].
 */

-(BOOL)Mle: (Interval *)value {
  double vmin, vmax;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: NULL min: &vmin max: &vmax];

  retval = possibly_le(min, max, vmin, vmax);

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}

/* Mlt: value -> Boolean [POP]
 *
 * Returns true iff there exists a member of [self->min, self->max]
 * that is less than a member of [value->min, value->max].
 */

-(BOOL)Mlt: (Interval *)value {
  double vmin, vmax;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: NULL min: &vmin max: &vmax];

  retval = possibly_lt(min, max, vmin, vmax);

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}

/* Mge: value -> Boolean [POP]
 *
 * Returns true iff there exists a member of [self->min, self->max]
 * that is greater than or equal to a member of [value->min, value->max].
 */

-(BOOL)Mge: (Interval *)value {
  double vmin, vmax;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: NULL min: &vmin max: &vmax];

  retval = possibly_ge(min, max, vmin, vmax);

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}

/* Leq: value -> Boolean [POP]
 *
 * Returns true iff all members of [self->min, self->max] are equal to
 * all members of [value->min, value->max]. Clearly, this is only when
 * each set contains a single equally-valued member. 
 */

-(BOOL)Leq: (Interval *)value {
  double vmin, vmax;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: NULL min: &vmin max: &vmax];

  retval = necessarily_eq(min, max, vmin, vmax);

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}

/* Lne: value -> Boolean [POP]
 *
 * Returns true iff all members of [self->min, self->max] are not
 * equal to any member of [value->min, value->max].
 */

-(BOOL)Lne: (Interval *)value {
  double vmin, vmax;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: NULL min: &vmin max: &vmax];

  retval = necessarily_ne(min, max, vmin, vmax);

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}

/* Lgt: value -> Boolean [POP]
 *
 * Returns true iff all members of [self->min, self->max] are greater
 * than all members of [value->min, value->max]. 
 */

-(BOOL)Lgt: (Interval *)value {
  double vmin, vmax;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: NULL min: &vmin max: &vmax];

  retval = necessarily_gt(min, max, vmin, vmax);

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}

/* Lle: value -> Boolean [POP]
 *
 * Returns true iff all members of [self->min, self->max] are less
 * than or equal to all members of [value->min, value->max]. 
 */

-(BOOL)Lle: (Interval *)value {
  double vmin, vmax;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: NULL min: &vmin max: &vmax];

  retval = necessarily_le(min, max, vmin, vmax);

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}

/* Llt: value -> Boolean [POP]
 *
 * Returns true iff all members of [self->min, self->max] are less
 * than all members of [value->min, value->max]. 
 */

-(BOOL)Llt: (Interval *)value {
  double vmin, vmax;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: NULL min: &vmin max: &vmax];

  retval = necessarily_lt(min, max, vmin, vmax);

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}

/* Lge: value -> Boolean [POP]
 *
 * Returns true iff all members of [self->min, self->max] are greater
 * than or equal to all members of [value->min, value->max]. 
 */

-(BOOL)Lge: (Interval *)value {
  double vmin, vmax;
  BOOL retval;

  CHECK_DEF_ARG(value);

  [value get: NULL min: &vmin max: &vmax];

  retval = necessarily_ge(min, max, vmin, vmax);

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}

/* contains: value -> Boolean [POP]
 *
 * Returns true iff the set of numbers belonging to the interval value is
 * a subset of the set of numbers belonging to the interval self.
 */

-(BOOL)contains: (Interval *)value {
  double vmin, vmax;
  BOOL retval;

  [value get: NULL min: &vmin max: &vmax];

  retval = (min <= vmin && max >= vmax) ? YES : NO;

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}

/* intersects: value -> Boolean [POP]
 *
 * Returns true iff the set of numbers belonging to the interval value
 * intersects with the selt of numbers belonging to the interval self.
 */

-(BOOL)intersects: (Interval *)value {
  double vmin, vmax;
  BOOL retval;

  [value get: NULL min: &vmin max: &vmax];

  retval = (necessarily_gt(min, max, vmin, vmax)
	    || necessarily_lt(min, max, vmin, vmax)) ? NO : YES;

  IF_STACK_TOP(value) [pop() drop];
  IF_STACK_TOP(self) [pop() drop];

  return retval;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Contextual methods for stack management. */

/* ret -> self
 *
 * This method is designed to handle the stack when returning a value
 * from a function. The value will either be a local variable or the
 * result of an expression. In the former case, it needs to be pushed
 * onto the stack so it can be dropped later in the expression
 * containing the function call. In the latter case, nothing needs to
 * be done as the variable is already on the stack.
 *
 * If the return value is a global or instance variable, the return
 * expression should be return [[var clone] ret].
 */

-ret {
  IF_STACK_TOP(self) return self;
  else return push(self);
}

/* call -> Interval
 *
 * This method is designed to handle the stack when using a value as a
 * call-by-value function call. The value will either be a variable
 * scoped within the calling statement block, or the result of an
 * expression. In the former case, the variable should be cloned so it
 * can be passed as an argument to the function. In the latter case it
 * should be taken off the stack. The arguments to a function are
 * local to that function, and should all be dropped before they go
 * out of scope.
 */

-call {
  IF_STACK_TOP(self) return pop();
  else return [self clone];
}

/* fumble -> self
 *
 * This method is designed to handle cases when a local variable is
 * used as part of an expression in a return statement. The ret method
 * could in fact equally be used under these circumstances, but the
 * original idea was to have just one ret call for each return
 * statement. Fumbling is what happens immediately prior to something
 * being dropped, so the idea is that a local variable can be fumbled
 * by putting it onto the stack, whereupon it will be dropped by any
 * enclosing method call in the expression.
 */

-fumble {
  return push(self);
}

/* castVoid [POP]
 *
 * This method is used when an interval is returned from a function but
 * the result is unused by the caller.
 */

-(void)castVoid {
  IF_STACK_TOP(self) [pop() drop];
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Destruction method */

/* drop
 *
 * Destroy the Interval, maintaining the count of currently created
 * Intervals.
 */

-(void)drop {
# ifdef TRACE
  printf("Dropped %p [%g, %g, %g]\n", self, min, var, max);
# endif
  TRACE_SELF;
  nintervals--;
  [super drop];
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Information methods */

/* gap -> max - min
 *
 * Return the size of the gap between minimum and maximum in the interval.
 */

-(double)gap {
  return max - min;
}

@end
